'''
Code that is necessary for supporting older Django versions
'''

try:  # Added in 1.7
    from django.apps import apps
    get_model = apps.get_model
except ImportError:  # Django < 1.7 pragma: no cover
    from django.db.models import get_model

try:  # Added in 1.7
    from django.utils.module_loading import import_string as import_by_path
except ImportError:  # Django < 1.7 pragma: no cover
    def import_by_path(dotted_path, error_prefix=''):  # pragma: no cover
        """
        Added in Django 1.6 to django.utils.module_loading

        Import a dotted module path and return the attribute/class designated by the
        last name in the path. Raise ImproperlyConfigured if something goes wrong.
        """
        from importlib import import_module
        import sys

        from django.core.exceptions import ImproperlyConfigured
        from django.utils import six

        try:
            module_path, class_name = dotted_path.rsplit('.', 1)
        except ValueError:
            raise ImproperlyConfigured("%s%s doesn't look like a module path" % (
                error_prefix, dotted_path))
        try:
            module = import_module(module_path)
        except ImportError as e:
            msg = '%sError importing module %s: "%s"' % (
                error_prefix, module_path, e)
            six.reraise(ImproperlyConfigured, ImproperlyConfigured(msg),
                        sys.exc_info()[2])
        try:
            attr = getattr(module, class_name)
        except AttributeError:
            raise ImproperlyConfigured('%sModule "%s" does not define a "%s" attribute/class' % (
                error_prefix, module_path, class_name))
        return attr



import six

from PIL import Image

from django.core.files.base import ContentFile
import django.core.files


class BasePILTransform(object):
    '''
    Base transform object that provides helper methods to transform
    :py:class:`django.core.files.images.ImageFile` using
    :py:mod:`PIL`.

    Must subclass and override :py:meth:`~.BasePILTransform.transform_pil_image`.
    '''

    IMAGE_QUALITY = 85

    def __call__(self, original_django_file):
        '''
        Returns the transformed version of :py:class:`PIL.Image.Image`

        Uses :py:meth:`~.BasePILTransform.transform_pil_image` to transform
        the :py:class:`PIL.Image.Image`.

        :param original_django_file: source file
        :type original_django_file: :py:class:`django.core.files.images.ImageFile`
        :return: transformed file
        :rtype: :py:class:`django.core.files.File`
        '''
        if not isinstance(original_django_file, django.core.files.File):
            raise TypeError(
                'image is a {0}, not a django File'.format(type(original_django_file))
            )
        if not original_django_file.file:
            raise ValueError(
                'the django file has no file saved to it.'
            )

        return self.pil_image_to_django_file(
            self.transform_pil_image(
                self.django_file_to_pil_image(original_django_file)
            )
        )

    def pil_image_to_django_file(self, pil_image):
        '''
        Gets a PIL image ready to be able to be saved using
        :py:meth:`django.db.models.fields.files.FieldFile.save`

        It converts the mode first to ``RGB`` or ``L``, so that it can
        then save it as a ``JPEG``. It will save it as a progressive
        ``JPEG`` with a quality of :py:attr:`IMAGE_QUALITY`.

        :param pil_image: original image
        :type pil_image: :py:class:`PIL.Image.Image`
        :return: transformed image
        :rtype: :py:class:`django.core.files.base.ContentFile`
        '''
        if pil_image.mode not in ('L', 'RGB'):
            pil_image = pil_image.convert("RGB")
        temp_io = six.BytesIO()

        pil_image.save(
            temp_io,
            "JPEG",
            quality=self.IMAGE_QUALITY,
            optimize=True,
            progressive=True
        )

        temp_io.seek(0)
        django_file = ContentFile(temp_io.getvalue())
        return django_file

    def transform_pil_image(self, pil_image):
        '''
        Returns the transformed version of the :py:class:`PIL.Image.Image`
        Do some logic on :py:class:`PIL.Image.Image`.

        Must subclass method to provide transformation logic.

        :param pil_image: original image
        :type pil_image: :py:class:`PIL.Image.Image`
        :return: transformed image
        :rtype: :py:class:`PIL.Image.Image`
        '''
        raise NotImplementedError

    def django_file_to_pil_image(self, django_file):
        '''
        Converts a the file returned by
        :py:class:`django.db.models.fields.ImageField` to a PIL image.

        :param django_file: django file
        :type django_file: :py:class:`django.db.models.fields.files.FieldFile`
        :rtype: :py:class:`PIL.Image.Image`
        '''
        django_file.open()
        pil_image = Image.open(django_file.file)
        return pil_image


class Scale(BasePILTransform):
    '''
    Scales down an image to max height and/or width. If the original
    image is smaller than both/either specified dimensions than it will
    be left unchanged.
    '''

    def __init__(self, width=None, height=None):
        '''
        Initialize this class with a max height and/or width (in pixels).

        :param height: max height of scaled image
        :param width: max width of scaled image
        :type height: int or float
        :type width: int or float
        '''
        self.dimensions = (width, height)
        if not any(self.dimensions):
            raise ValueError(
                'Must be called with either `height` or `width`'
            )

    def transform_pil_image(self, pil_image):
        '''
        Uses :py:meth:`PIL.Image.Image.transform` to scale
        down the image.

        Based on `this stackoverflow discussions <http://stackoverflow.com/a/940368/907060>`_, uses
        :attr:`PIL.Image.ANTIALIAS`
        '''
        max_width = min(self.dimensions[0] or float('inf'), pil_image.size[0])
        max_height = min(self.dimensions[1] or float('inf'), pil_image.size[1])
        max_dimensions = (max_width, max_height)

        pil_image.thumbnail(max_dimensions, Image.ANTIALIAS)
        return pil_image



import os


from simpleimages.django_compat import import_by_path


def get_caller():
    from django.conf import settings
    caller_text = getattr(
        settings,
        'SIMPLEIMAGES_TRANSFORM_CALLER',
        'simpleimages.callers.default'
    )
    return import_by_path(caller_text)


def perform_transformation(instance, field_names_to_transform=None):
    '''
    Transforms a model based on the fields specified in the
    ``transformed_fields`` attribute. This should map source image
    field names to dictionaries mapping destination field name to
    their transformations. For instance::

        {
            'image': {
                'thumbnail': scale(width=10),
            }
        }

    If ``field_names_to_transform`` is None, then it will transform
    all fields. Otherwise it will only transform
    from those fields specified in ``field_names_to_transform``.

    :param instance: model instance to perform transformations on
    :type instance: instance of :py:class:`django.db.models.Model`

    :param field_names_to_transform: field names on model to perform transformations on
    :type field_names_to_transform: iterable of strings or None
    '''
    for source_field_name, destination_dict in instance.transformed_fields.items():
        if field_names_to_transform is None or source_field_name in field_names_to_transform:
            for destination_field_name, transformation in destination_dict.items():
                arguments = [instance, source_field_name, destination_field_name, transformation]
                get_caller()(transform_field, *arguments)


def transform_field(instance, source_field_name, destination_field_name, transformation):
    '''
    Does an image transformation on a instance. It will get the image
    from the source field attribute of the instnace, then call
    the transformation function with that instance, and finally
    save that transformed image into the destination field attribute
    of the instance.

    .. note::

        If the source field is blank or the transformation returns
        a false value then the destination field image will be deleted, if it
        exists.

    .. warning::

        When the model instance is saved with the new transformed image, it uses
        the ``update_fields`` argument for
        :py:meth:`~django.db.models.Model.save`, to tell the model to only update
        the destination field and, if set in the destination field, the
        :py:attr:`~django.db.models.ImageField.height_field` and
        :py:attr:`~django.db.models.ImageField.width_field`. This means that
        if the saving code for the model sets any other fields, in the saving
        field process, it will not save those fields to the database. This would
        only happen if you introduce custom logic to the saving process of
        destination field, like the dimension fields do, that updates another field
        on that module. In that case, when the model is saved for the
        transformation, that other field will not be saved to the database.


    :param instance: model instance to perform transformations on
    :type instance: instance of :py:class:`django.db.models.Model`

    :param source_field_name: field name on model to find source image
    :type source_field_name: string

    :param destination_field_name: field name on model save transformed image to
    :type destination_field_name: string

    :param transformation: function, such as :py:func:`~.transforms.scale`, that takes an image files and returns a transformed image
    :type transformation: function
    '''

    source_field = getattr(instance, source_field_name)
    destination_field = getattr(instance, destination_field_name)
    update_fields = [destination_field_name]
    transformed_image = get_transformed_image(source_field, transformation)
    if transformed_image:
        destination_name = os.path.basename(source_field.name)
        dimension_field_names = [
            destination_field.field.height_field,
            destination_field.field.width_field]
        update_fields += filter(None, dimension_field_names)
        destination_field.save(
            destination_name,
            transformed_image,
            save=False
        )
    elif destination_field:
        destination_field.delete()
    else:
        return
    instance.save(update_fields=update_fields)


def get_transformed_image(source, transformation):
    if source:
        source.open()
        return transformation(source)



from django.db.models.signals import post_save
from django.dispatch import receiver

import simpleimages.utils


def track_model(model):
    '''
    Perform designated transformations on model, when it saves.

    Calls :py:func:`~simpleimages.utils.perform_transformation`
    on every model saves using
    :py:data:`django.db.models.signals.post_save`.

    It uses the ``update_fields`` kwarg to tell what fields it should
    transform.
    '''
    @receiver(post_save, sender=model, weak=False, dispatch_uid='simpleimages')
    def transform_signal(sender, **kwargs):
        simpleimages.utils.perform_transformation(
            kwargs['instance'],
            kwargs['update_fields']
        )

    def disconnect():
        post_save.disconnect(sender=model, dispatch_uid='simpleimages')
    return disconnect



from clint.textui import puts, indent, progress

from django.core.management.base import BaseCommand

import simpleimages.utils
from simpleimages.django_compat import get_model


def parse_model_specifier(specifier):
    '''
    Parses a string that specifies either a model or a field.
    The string should look like ``app.model.[field]``.

    >>> print parse_model_specifier('tests.TestModel')
    (<class 'tests.models.TestModel'>, None)
    >>> print parse_model_specifier('tests.TestModel.image')
    (<class 'tests.models.TestModel'>, 'image')

    :return: model and (optionally) field name
    :rtype: tuple of :py:class:`~django.db.models.Model` and str or None
    '''
    values = specifier.split('.')

    if len(values) == 2:
        values.append(None)
    elif len(values) != 3:
        raise ValueError(
            'Model specifier must be in app.model.[field] format. It'
            'has {} parts instead of 2 or 3 (when split on ".")'.format(
                len(values)
            )
        )

    app_name, model_name, field_name = values
    model = get_model(app_name, model_name)
    if not model:
        raise ValueError(
            'Model {} on app {} can not be found'.format(
                model_name,
                app_name,
            )
        )
    return model, field_name


class Command(BaseCommand):
    args = '[<app.model.[field], app.model.[field], ...>]'
    help = 'Retransforms imagefields for the model(s)'

    def handle(self, *args, **options):
        for arg in args:
            puts('Transforming {0}'.format(arg))
            model, field_name = parse_model_specifier(arg)

            instances = model._default_manager.all()
            number_instances = instances.count()
            with indent(4):
                if not number_instances:
                    puts('No instances found')
                    continue

                else:
                    puts('{0} models found'.format(instances.count()))
                if field_name:
                    puts('From source field name "{0}"'.format(field_name))
            field_names = field_name or None

            for instance in progress.bar(instances):
                simpleimages.utils.perform_transformation(instance, field_names)



def default(function, *args, **kwargs):
    '''
    Calls ``function`` with any passed in ``args`` and ``kwargs.
    '''
    function(*args, **kwargs)


def _no_action(function, *args, **kwargs):
    '''
    Does nothing. For testing.
    '''
    pass


def celery(function, *args, **kwargs):
    '''
    Calls ``function`` asynchronously by creating a
    `shared-task <http://docs.celeryproject.org/en/latest/django/first-steps-with-django.html#using-the-shared-task-decorator>`_
    with celery.
    '''
    from celery import shared_task
    shared_task(function).delay(*args, **kwargs)



from setuptools import setup


setup(
    name='django-simpleimages',
    version='1.3.2',
    author='Saul Shanabrook',
    author_email='s.shanabrook@gmail.com',
    packages=[
        'simpleimages',
        'simpleimages.management',
        'simpleimages.management.commands',
    ],
    url='https://www.github.com/saulshanabrook/django-simpleimages',
    license=open('LICENSE.txt').read(),
    description='Opinionated Django image transforms on models',
    long_description=open('README.rst').read(),
    install_requires=[
        "Django>=1.5,<1.10",
        "six",
        "Pillow",
        'clint',
    ],
    zip_safe=False,  # so that django finds management commands,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Framework :: Django',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries',
    ],
)
